
create or replace function hwp.watchlist(
    iplate varchar(128),
    idistance int default 0
) returns jsonb as $$
    select lib.jstripb(jsonb_agg(r)) from (
    select 
        t.owner as "list-owner",
        t.ident as "list-ident",
        t.utime as "list-date",
        t.description as "list-description",
        l.plate,
        case when $2 > 0 then lib.ilevenstein(l.plate::text,$1::text) else 0 end as "plate-distance",
        l.class, l.color, l.make, l.model, l.alarm, l.memo, l.meta
    from only hwp.watchlistitems l 
        join hwp.watchlist t 
        on l.owner = t.owner and l.ident = t.ident
    where 
        case when $2 > 0 then lib.ilevenstein(l.plate::text, $1::text) <= $2 else l.plate = $1 end
    order by "plate-distance" ) r;    
$$ language sql;


create or replace function hwp.listwatchlist (
    iowner  varchar(128),
    iident varchar(128)
) returns jsonb as $$    
    select 
        lib.jstripb(jsonb_object_agg(owner,lists))
    from(
        select 
            owner,json_agg(r.*) as lists 
        from 
            hwp.watchlist r
        where 
            ($1 = '*' or ($1 ~ '[\*]' and owner ~ ('^'||replace($1,'*','.*')||'$')) or owner = $1) and 
            ($2 = '*' or ($2 ~ '[\*]' and ident ~ ('^'||replace($2,'*','.*')||'$')) or ident = $2)
        group by owner
    ) q;
$$ language sql;


create or replace function hwp.listwatchlistitems (
    iowner  varchar(128),
    iident varchar(128)
) returns jsonb as $$    
    select lib.jstripb(jsonb_object_agg(owner,lists)) 
    from (
        select 
            owner,json_object_agg(ident,items) as lists 
        from(
            select owner,ident,json_agg(
                json_build_object(
                    'plate',i.plate,
                    'class',i.class,
                    'color',i.color,
                    'make',i.make,
                    'model',i.model,
                    'alarm',i.alarm,
                    'memo',i.memo,
                    'meta',i.meta
                )) as items 
            from 
                hwp.watchlistitems i join hwp.watchlist l using(owner,ident) 
                where 
                    ($1 = '*' or ($1 ~ '[\*]' and owner ~ ('^'||replace($1,'*','.*')||'$')) or owner = $1) and 
                    ($2 = '*' or ($2 ~ '[\*]' and ident ~ ('^'||replace($2,'*','.*')||'$')) or ident = $2)
                group by owner,ident
            ) q  
            group by owner
        ) qq;
$$ language sql;

create or replace function hwp.watchlist_delete(
    iowner  varchar(128),
    iident varchar(128),
    out affected integer
) as $$
begin
    delete from hwp.watchlistlog where (iowner = '*' or owner = iowner) and (iident = '*' or ident = iident);
    delete from hwp.watchlistvcs where (iowner = '*' or owner = iowner) and (iident = '*' or ident = iident);
    delete from hwp.watchlistitems where (iowner = '*' or owner = iowner) and (iident = '*' or ident = iident);
    delete from hwp.watchlist where (iowner = '*' or owner = iowner) and (iident = '*' or ident = iident);
    get diagnostics affected = ROW_COUNT;
    return;
end;
$$ language plpgsql;


create or replace function hwp.watchlist_cleanup(
    iowner  varchar(128),
    iident varchar(128),
    imoment timestamptz default null,
    irevision varchar(32) default null,
    ihash varchar(40) default null,
    out affected integer
) as $$
declare
    _o varchar(128);
    _i varchar(128);
    _r varchar(32);
    _d timestamptz;
    r record;
begin
    affected := 0;
    for _o in select distinct owner from hwp.watchlist where iowner = '*' or owner = iowner loop
        for _i in select distinct ident from hwp.watchlist where owner = _o and (iident = '*' or ident = iident) loop
            if irevision notnull then
                select into _r,_d revision,moment from hwp.watchlistlog where owner = _o and ident = _i and revision = irevision; 
            elsif ihash notnull then
                select into _r,_d revision from hwp.watchlistlog where owner = _o and ident = _i and sha1 = ihash order by revision limit 1;
            elsif imoment notnull then
                select into _r,_d revision,moment from hwp.watchlistlog where owner = _o and ident = _i and moment >= imoment order by revision limit 1;
            else
                select into _r,_d revision,moment from hwp.watchlistlog where owner = _o and ident = _i order by revision desc limit 1;
            end if;
            if not found then
                return;
            end if;
            drop table if exists t;
            create temporary table t (like hwp.watchlistitems) on commit drop;
            for r in select * from hwp.watchlistvcs where  owner = _o and ident = _i and revision <= _r order by revision loop
                if r.op = 1 then
                    insert into t (owner,ident,plate,class,color,make,model,alarm,memo,meta) values(r.owner,r.ident,r.plate,r.class,r.color,r.make,r.model,r.alarm,r.memo,r.meta);
                elsif r.op = 0 then
                    update t set class=r.class,color=r.color,make=r.make,model=r.model,alarm=r.alarm,memo=r.memo,meta=r.meta where plate = r.plate;
                elsif r.op = -1 then
                    delete from t where plate = r.plate;
                end if;
            end loop;
            delete from hwp.watchlistvcs where  owner = _o and ident = _i and revision <= _r;
            get diagnostics affected = ROW_COUNT;
            insert into hwp.watchlistvcs (op,optime,revision,sha1,owner,ident,plate,class,color,make,model,alarm,memo,meta)  
            select 
                1 as op, _d as optime, _r as revision, ihash as sha1, iowner,ident,plate,class,color,make,model,alarm,memo,meta
                from t;
            delete from hwp.watchlistlog where owner = _o and ident = _i and revision <= _r;
            insert into hwp.watchlistlog (moment,owner,ident,revision,sha1,inserted) 
            select
                _d as moment, iowner as owner, iident as ident, _r as revision, ihash as sha1, affected as inserted;
        end loop;
    end loop;
    return;
end;    
$$ language plpgsql;

create or replace function hwp.watchlist_revert(
    iowner  varchar(128),
    iident varchar(128),
    ioffset int default 1,
    irevision varchar(32) default null,
    out inserted integer,
    out updated integer,
    out deleted integer
) as $$
declare 
    r record;
    rr record;
    p varchar(32);
    c integer;
    cc integer;
begin
    c := 0;
    inserted := 0;
    updated := 0;
    deleted := 0;
    if irevision isnull and (ioffset isnull or ioffset < 1 ) then
       raise exception 'Invalid offset' using hint='401';
    end if;   
    select into p replace(ct.uuid_generate_v1() :: text, '-', '');
    execute format('create temp table %I (like hwp.watchlistvcs) on commit drop',p);
    drop table if exists olog;
    create temp table olog (like hwp.watchlistlog) on commit drop;
    if irevision notnull then
        if exists (select 1 from hwp.watchlistlog where owner = iowner and ident = iident and revision = irevision) then
            insert into olog select * from hwp.watchlistlog where owner = iowner and ident = iident and revision > irevision;
        end if;
    else
        insert into olog select * from hwp.watchlistlog where owner = iowner and ident = iident order by revision desc limit ioffset;
    end if;    
    for r in select * from olog order by revision desc loop
        execute format('insert into %I 
            select 
                -op as op
                ,optime,revision,sha1,owner,ident,plate,class,color,make,model,alarm,memo,meta
            from hwp.watchlistvcs
            where owner = $1 and ident = $2 and revision = $3
            ', p) using iowner, iident, r.revision;
       delete from hwp.watchlistvcs where revision = r.revision;
       delete from hwp.watchlistlog where revision = r.revision;
       get diagnostics cc = ROW_COUNT;
       c := c + cc;
    end loop;
    if c > 0 then
        select into inserted,deleted,updated * from hwp.watchsweep (p, null, null, false);
    end if;    
end;
$$ language plpgsql;

