
--
-- { "object":"origin", "action":"discovery", "format":["raw","public"]}
--
-- { "object":"lane", "action":"discovery",... "origin": }
--
-- { "object":"profile", "action":"discovery" }
--
-- { "object":"string", "action":"discovery",... "section":..."lang":... }
--
-- { "object":"opt","action":"discovery",...,"section":...,"keys":...,"lang":... }
--
-- { "object":"const","action":"discovery","kind":["int","text"],"section",...}
--
create or replace function api_discovery (idata jsonb, out errcode integer, out errmsg varchar, out result jsonb) as $$
declare
	_object text;
    _post uuid;
    _channel integer;
    _section text;
    _keys text;
    _kind text;
    _format text;
    _origin uuid;
    _lang varchar(2);
begin
    errcode := 0;
    errmsg := '';
    _object := idata ->> 'object';
    if _object = 'origin' then
        _format := idata ->> 'format';
        if _format isnull or _format = 'raw' then
            select 
                into result 
                jsonb_agg(row_to_json(r)) 
                from ( 
                    select * 
                    from collection_event_origin order by origin->>'time' desc ) r;
        elsif _format = 'public' then
            select 
                into result 
                jsonb_agg(jsonb_build_object(
                    'id', id,
                    'time', origin ->> 'time',
                    'name', name,
                    'services', param -> 'public' -> 'services'
                )) 
                from ( 
                    select * 
                    from collection_event_origin order by origin->>'time' desc ) r;
        elsif _format = 'param' then
            select 
                into result 
                lib.jstripb(jsonb_agg(param)) 
                from (
                    select * 
                    from collection_event_origin order by origin->>'time' desc ) r;
        elsif _format = 'param-public' then
            select 
                into result 
                lib.jstripb(jsonb_agg(param->'public')) 
                from (
                    select * 
                    from collection_event_origin order by origin->>'time' desc ) r;
        else
            errcode := -2;
            errmsg = "Invalid format";
            return;
        end if;
    elsif _object = 'lane' then
        _origin := (idata ->> 'origin')::uuid;
        select 
            into result    
            jsonb_object_agg(origin,lanes) as lanes 
            from (
                select 
                    origin, jsonb_agg(lane) as lanes 
                from 
                    collection_event_lane 
                where _origin isnull or origin = _origin
                group by origin) r;
    elsif _object = 'profile' then
        _origin := (idata ->> 'origin');
        select
            into result 
            jsonb_object_agg(type,profiles)
            from (
                select 
                    doctype as type, jsonb_agg(profile) as profiles 
                from 
                    collection_event_profile 
                group by doctype    
            ) r;
    elsif _object = 'string' then
        _section := (idata ->> 'section');
        _lang := coalesce((idata ->> 'lang'),(_o('language_priority')::text[])[1]);
        select into result
            jsonb_agg(r) 
            from (
                select 
                    key as value, 
                    value as caption 
                from 
                    public.stringtable 
                where section = _section and lang = _lang
            ) r; 
    elsif _object = 'opt' then
        _section := (idata ->> 'section');
        _keys := (idata ->> 'keys');
        _lang := coalesce((idata ->> 'lang'),(_o('language_priority')::text[])[1]);
        select into result 
            jsonb_agg(r)
            from (
                select 
                    o.section,o.key,o.value,o.default, o.type, o.group, o.weight,
                    s.value as description 
                from 
                    core.opt o 
                    left join public.stringtable s 
                        on s.section= 'opt' and 
                            s.lang = _lang and 
                            s.key = _section||'.'||o.key 
                where 
                    case 
                        when _section isnull or _section = '' 
                        then true
                        else o.section = _section 
                    end
                    and 
                    case 
                        when (_keys isnull or _keys = '') 
                        then true 
                        else o.key ~ _keys
                    end    
            ) r; 
    elsif _object = 'const'  then
        _section := (idata ->> 'section');
        _kind := (idata ->> 'kind');
		if _kind isnull or _kind = '' then
			select into result 
			row_to_json(rr)::jsonb 
			from ( 
                select
				( select json_object_agg(name, value)
					from public.const_integer
					where case when _section isnull or _section = '' then true else section = _section end
				) as const_int,	
				( select json_object_agg(name, value)
					from public.const_text
					where case when _section isnull or _section = '' then true else section = _section end 
				) as const_text	
			) rr;
		elsif _kind = any(array['int','integer']) then
			select into result 
            json_object_agg(name, value)
                from public.const_integer
                where section = _section; 
		elsif _kind = any(array['text','string']) then 
			select into result 
            json_object_agg(name, value)
                from public.const_text
                where section = _section; 
		else
            errcode := -2;
            errmsg := 'Invalid kind';
		end if;
    else
		errcode := -1;
        errmsg = 'Invalid object';
	end if;
exception when others then
    errcode := -10;
    errmsg := format('%s: %s',SQLSTATE,to_jsonb(SQLERRM));
end;
$$ language plpgsql;


--
-- { "object":"report", "action":"get" }
-- { ..., "name":"", "param":{...} }
--
create or replace function api_report_get (idata jsonb, out errcode integer, out errmsg varchar, out result jsonb) as $$
declare
    _name varchar  := idata ->> 'name';
    _param jsonb := coalesce(idata -> 'param', '{}'::jsonb);      
    _handler varchar;
begin
    errcode := 0;
    errmsg := '';
        if _name isnull then
            errcode := -1;
            errmsg := 'Invalid report name';
            return;
        end if;    
        select into _handler handler from public.handlers where section = 'report' and handle = _name and enabled;
        if not found then
            errcode := 1;
            errmsg := 'Valid report handler not found';
            return;
        end if;
        execute format('select * from %s ($1)', _handler) using _param into errcode, errmsg, result;
exception when others then
    errcode := -10;
    errmsg := format('%s: %s',SQLSTATE,to_jsonb(SQLERRM));
end;
$$ language plpgsql;

--
-- { "object":"event", "action":"add", "event": { ... } }
--

create or replace function api_event_add(idata jsonb, out errcode integer, out errmsg varchar, out result jsonb) as $$
declare
    _store public.eventstore;
begin
    errcode := 0;
    errmsg := '';
    _store := public._event_prepare(idata);
    if _store.id isnull or _store.scrollid isnull then 
        errcode := 2;
        errmsg := format('Invalid insert -> id: %s sid: %s',
            coalesce(_store.id::text,'<null>'), 
            coalesce(_store.scrollid::text,'<null>'));
    else
        select into errcode,errmsg * from public._event_postfactum(_store);
        result = json_build_object('id',_store.id,'scrollid',_store.scrollid)::jsonb;
    end if;
exception when others then
    errcode := -10;
    errmsg := format('%s: %s',SQLSTATE,to_jsonb(SQLERRM));
end;
$$ language plpgsql;


--
-- { "object":"event", "action":"clear", "store": "fromtime": "totime": }
--

create or replace function api_event_clear(idata jsonb, out errcode integer, out errmsg varchar, out result jsonb) as $$
declare
    _id bigint;
    istore varchar;
    ifromtime timestamptz;
    itotime timestamptz;
    _cnt integer;
begin 
    errcode := 0;
    errmsg := '';
    istore := idata ->> 'store';
    ifromtime := (idata ->> 'fromtime')::timestamptz;
    itotime   := (idata ->> 'totime')::timestamptz;
    lock table eventindex in share row exclusive mode;
    alter table eventindex disable trigger t_count_event;
    delete  
    from eventindex cascade
    where 
        case when istore isnull then true else store = istore end
        and ctime between 
            coalesce(ifromtime,'-infinity'::timestamptz) and
            coalesce(itotime,'infinity'::timestamptz) 
    ;    
    get diagnostics _cnt = ROW_COUNT;
    insert into core.relcounting values('public','eventindex',-_cnt);
    alter table eventindex enable trigger t_count_event;
    result = json_build_object('affected',_cnt)::jsonb;
exception when others then
    errcode := -10;
    errmsg := format('%s: %s',SQLSTATE,to_jsonb(SQLERRM));
end;
$$ language plpgsql;

--
-- { "object":"origin", "action":"set", "uuid":, "name":, "origin":, "param": }
--

create or replace function api_origin_set(idata jsonb, out errcode integer, out errmsg varchar, out result jsonb) as $$
declare
    iorigin uuid;
    iname varchar;
    ivalue jsonb;
    iparam jsonb;
begin
    errcode := 0;
    errmsg := '';
    iorigin := (idata ->> 'uuid')::uuid;
    if iorigin isnull then
        errcode := -2;
        errmsg := 'Invalid origin';
        return;
    end if;
    iname = idata ->> 'name';
    ivalue = idata -> 'origin';
    iparam = idata -> 'param';
    insert into public.collection_event_origin
        (id, name, origin, param)
    values ( 
        iorigin,
        iname,
        ivalue,
        iparam
    ) on conflict (id)
    do update
        set 
             name = coalesce(iname,public.collection_event_origin.name)
            ,origin = coalesce(ivalue,public.collection_event_origin.origin)
            ,param = coalesce(iparam,public.collection_event_origin.param)
        where public.collection_event_origin.id = iorigin
    ;
exception when others then
    errcode := -10;
    errmsg := format('%s: %s',SQLSTATE,to_jsonb(SQLERRM));
end;
$$ language plpgsql;

--
-- { "object":"value", "action":"set", "section":, "key":, "value": }
--

create or replace function api_value_set(idata jsonb, out errcode integer, out errmsg varchar, out result jsonb) as $$
declare
    isection varchar;
    ikey varchar;
    ivalue jsonb;
begin
    errcode := 0;
    errmsg := '';
    isection := idata ->> 'section';
    ikey := idata ->> 'key';
    if isection isnull or ikey isnull then
        errcode := -2;
        errmsg := 'Invalid section-key';
        return;
    end if;
    insert 
        into public.valuestore 
        (section,key,value) 
    values (
        isection,
        ikey,
        idata -> 'value'
    );
exception when others then
    errcode := -10;
    errmsg := format('%s: %s',SQLSTATE,to_jsonb(SQLERRM));
end;
$$ language plpgsql;

--
-- { "object":"value", "action":"get", "section":, "key": }
--

create or replace function api_value_get(idata jsonb, out errcode integer, out errmsg varchar, out result jsonb) as $$
declare
    isection varchar;
    ikey varchar;
begin
    errcode := 0;
    errmsg := '';
    isection := idata ->> 'section';
    ikey := idata ->> 'key';
    if isection isnull then
        errcode := -2;
        errmsg := 'Invalid section';
        return;
    end if;
    if ikey notnull then
        select into result 
        value from public.valuestore
        where section = isection and key = ikey;
    else
        select into result
            jsonb_object_agg(key,value)
        from public.valuestore
        where section = isection;
    end if;
exception when others then
    errcode := -10;
    errmsg := format('%s: %s',SQLSTATE,to_jsonb(SQLERRM));
end;
$$ language plpgsql;


--
-- { "object":"value", "action":"drop", "section":, "key":  }
--

create or replace function api_value_drop(idata jsonb, out errcode integer, out errmsg varchar, out result jsonb) as $$
declare
    isection varchar;
    ikey varchar;
begin
    errcode := 0;
    errmsg := '';
    isection := idata ->> 'section';
    ikey := idata ->> 'key';
    if isection isnull then
        errcode := -2;
        errmsg := 'Invalid section';
        return;
    end if;
    if not ikey ~ '[*^$?]' then
        ikey := '^'||ikey||'$';
    end if;    
    with k as (
        delete 
            from public.valuestore
        where 
            section = isection and
            case when ikey isnull then true else key ~ ikey end
        returning key 
    ) select into result jsonb_agg(key) from k;    
exception when others then
    errcode := -10;
    errmsg := format('%s: %s',SQLSTATE,to_jsonb(SQLERRM));
end;
$$ language plpgsql;


-- { "object":"value", "action":"drop", "section":, "key":  }
--

create or replace function api_value_drop(idata jsonb, out errcode integer, out errmsg varchar, out result jsonb) as $$
declare
    isection varchar;
    ikey varchar;
begin
    errcode := 0;
    errmsg := '';
    isection := idata ->> 'section';
    ikey := idata ->> 'key';
    if isection isnull then
        errcode := -2;
        errmsg := 'Invalid section';
        return;
    end if;
    if not ikey ~ '[*^$?]' then
        ikey := '^'||ikey||'$';
    end if;    
    with k as (
        delete 
            from public.valuestore
        where 
            section = isection and
            case when ikey isnull then true else key ~ ikey end
        returning key 
    ) select into result jsonb_agg(key) from k;    
exception when others then
    errcode := -10;
    errmsg := format('%s: %s',SQLSTATE,to_jsonb(SQLERRM));
end;
$$ language plpgsql;

